unit uMain;

interface

uses
  System.SysUtils, System.Types, System.UITypes, System.Rtti, System.Classes,
  System.Variants, System.UIConsts, System.DateUtils, FMX.Types, FMX.Controls, FMX.Forms, FMX.Dialogs,
  FMX.TabControl, FMX.Objects, FMX.Effects, FMX.Layers3D, FMX.Types3D,
  FMX.Layouts, FMX.Edit, FMX.Memo;

const
  GAME_SIZE = 6;

type
  TEditLogic = record
    DelayTime: Integer; { A jtk elkezdhes kpest mennyivel ksbb kezddik el (ezredmsodperc) }
    FadeInTime: Integer;  { Edit megjelens egy lpsnek ideje (ezredmsodperc) }
    FadeInStep: Double; { Edit mekkora lpskzzel lesz lthat }
    FadeOutTime: Integer; { Edit eltns egy lpsnek ideje (ezredmsodperc) }
    FadeOutStep: Double; { Edit mekkora lpskzzel rejtdik el }
    VisibleTime: Integer; { Edit mennyi ideig lthat (ezredmsodperc) }
    Timer: TTimer;
  end;

  TGameMatrix = array[1..GAME_SIZE,1..GAME_SIZE] of Boolean;
  TPositionMatrix = array[1..GAME_SIZE,1..GAME_SIZE] of TPosition;

  TfrmMain = class(TForm)
    CalloutPanel1: TCalloutPanel;
    tcMain: TTabControl;
    tiMenu: TTabItem;
    tiGame: TTabItem;
    adDifficulty: TArcDial;
    lbDifficulty: TLabel;
    imgMain: TImage;
    btnStart: TButton;
    geDiff1: TGlowEffect;
    GlowEffect2: TGlowEffect;
    GlowEffect3: TGlowEffect;
    CalloutPanel2: TCalloutPanel;
    ShadowEffect1: TShadowEffect;
    geDiff2: TGlowEffect;
    geDiff3: TGlowEffect;
    geDiff4: TGlowEffect;
    layMain: TLayout;
    Image1: TImage;
    CalloutPanel3: TCalloutPanel;
    ShadowEffect2: TShadowEffect;
    pnlGame: TPanel;
    Panel2: TPanel;
    lbRed: TLabel;
    lbTimeText: TLabel;
    lbLifeText: TLabel;
    lbLife: TLabel;
    lbTime: TLabel;
    Panel3: TPanel;
    btnEnd: TButton;
    GlowEffect1: TGlowEffect;
    tmrTime: TTimer;
    adLength: TArcDial;
    GlowEffect4: TGlowEffect;
    lbLength: TLabel;
    geLength1: TGlowEffect;
    geLength2: TGlowEffect;
    geLength3: TGlowEffect;
    geLength4: TGlowEffect;
    GlowEffect5: TGlowEffect;
    lbGreen: TLabel;
    GlowEffect6: TGlowEffect;
    lbBlue: TLabel;
    GlowEffect7: TGlowEffect;
    edRed: TEdit;
    GlowEffect8: TGlowEffect;
    edGreen: TEdit;
    GlowEffect11: TGlowEffect;
    edBlue: TEdit;
    GlowEffect13: TGlowEffect;
    tmrRed: TTimer;
    tmrGreen: TTimer;
    tmrBlue: TTimer;
    memDebug: TMemo;
    lbStatus: TLabel;
    Label1: TLabel;
    procedure FormCreate(Sender: TObject);
    procedure adDifficultyChange(Sender: TObject);
    procedure btnStartClick(Sender: TObject);
    procedure tmrTimeTimer(Sender: TObject);
    procedure adLengthChange(Sender: TObject);
    procedure edRedClick(Sender: TObject);
    procedure edRedEnter(Sender: TObject);
    procedure edRedExit(Sender: TObject);
    procedure tmrRedTimer(Sender: TObject);

  private
    { Private declarations }
    FDifficulty: Integer;  { A nehzsg belltsra }
    FGameLength: Integer;  { A jtkhossz belltsra }
    FLife: Integer;  { Ennyiszer lehet rontani }
    FStartTime: TDateTime;  { A jtk ideje }
    FWordLength: Integer;  { A generlt szavak hossza }
    FGameMatrix: TGameMatrix;  { Az Edit-ek hol vannak ppen }

    FRedLogic: TEditLogic;
    FGreenLogic: TEditLogic;
    FBlueLogic: TEditLogic;
    FRotationDeltaMin: Integer;  { Min. hny fokot lehet fordtva az edit }
    FRotationDeltaMax: Integer;  { Max. hny fokot lehet fordtva az edit }

    FViewport: TViewport3D;  { A forgats animcihoz }
    FContainer: TLayer3D;  { A forgats animcihoz }

    procedure SwitchTo3D;
    procedure SwitchTo2D;
    procedure StartGame;
    procedure EndGame;
    procedure UpdateTime;
    procedure UpdateLife;
    procedure SetGameLogic;
    procedure UpdateEdit(AEdit: TEdit; ALogic: TEditLogic; AIsInit: Boolean = False);
    procedure Generate(AEdit: TEdit; ALogic: TEditLogic; AIsInit: Boolean = False);
    procedure ClearGameMatrix;
    procedure SetGameMatrix(var X: Integer; var Y: Integer); overload;
    procedure SetGameMatrix(AControl: TControl; var X: Integer; var Y: Integer); overload;
    procedure ResetControls;
    procedure GameOver(AIsOK: Boolean);
    procedure CheckLife(AEdit: TEdit);
  public
    { Public declarations }
  end;

var
  frmMain: TfrmMain;

implementation

{$R *.fmx}

procedure TfrmMain.adDifficultyChange(Sender: TObject);
begin
  { A nehzsg belltsa a felletrl }
  if adDifficulty.Value = 0 then
  begin
    lbDifficulty.Text:= 'Knny';
    geDiff1.Enabled:= True;
    geDiff2.Enabled:= False;
    geDiff3.Enabled:= False;
    geDiff4.Enabled:= False;
    FDifficulty:= 1;
  end
  else if adDifficulty.Value = -90 then
  begin
    lbDifficulty.Text:= 'Norml';
    geDiff1.Enabled:= False;
    geDiff2.Enabled:= True;
    geDiff3.Enabled:= False;
    geDiff4.Enabled:= False;
    FDifficulty:= 2;
  end
  else if (adDifficulty.Value = 180) or (adDifficulty.Value = -180) then
  begin
    lbDifficulty.Text:= 'Kicsit nehz';
    geDiff1.Enabled:= False;
    geDiff2.Enabled:= False;
    geDiff3.Enabled:= True;
    geDiff4.Enabled:= False;
    FDifficulty:= 3;
  end
  else if adDifficulty.Value = 90 then
  begin
    lbDifficulty.Text:= 'Nagyon nehz';
    geDiff1.Enabled:= False;
    geDiff2.Enabled:= False;
    geDiff3.Enabled:= False;
    geDiff4.Enabled:= True;
    FDifficulty:= 4;
  end;
end;

procedure TfrmMain.adLengthChange(Sender: TObject);
begin
  { A jtkid belltsa a felletrl }
  if adLength.Value = 0 then
  begin
    lbLength.Text:= '1 perc';
    geLength1.Enabled:= True;
    geLength2.Enabled:= False;
    geLength3.Enabled:= False;
    geLength4.Enabled:= False;
    FGameLength:= 1;
  end
  else if adLength.Value = -90 then
  begin
    lbLength.Text:= '2 perc';
    geLength1.Enabled:= False;
    geLength2.Enabled:= True;
    geLength3.Enabled:= False;
    geLength4.Enabled:= False;
    FGameLength:= 2;
  end
  else if (adLength.Value = 180) or (adLength.Value = -180) then
  begin
    lbLength.Text:= '3 perc';
    geLength1.Enabled:= False;
    geLength2.Enabled:= False;
    geLength3.Enabled:= True;
    geLength4.Enabled:= False;
    FGameLength:= 3;
  end
  else if adLength.Value = 90 then
  begin
    lbLength.Text:= '4 perc';
    geLength1.Enabled:= False;
    geLength2.Enabled:= False;
    geLength3.Enabled:= False;
    geLength4.Enabled:= True;
    FGameLength:= 4;
  end;
end;

procedure TfrmMain.GameOver(AIsOK: Boolean);
begin
  { A jtk vgn hvdik meg }
  tmrTime.Enabled:= False;
  tmrRed.Enabled:= False;
  tmrGreen.Enabled:= False;
  tmrBlue.Enabled:= False;

  if AIsOK then
    lbStatus.Text:= 'Sikerlt! :)'
  else
    lbStatus.Text:= 'Nem sikerlt! :(';

  lbStatus.Visible:= True;
  lbLifeText.Visible:= False;
  lbTimeText.Visible:= False;
  lbLife.Visible:= False;
  lbTime.Visible:= False;
end;

procedure TfrmMain.CheckLife(AEdit: TEdit);
begin
  { A bert szveg ellenrzse, hogy megegyezik-e az elvrt szveggel }
  if AEdit = edRed then
  begin
    if edRed.Text <> lbRed.Text then
    begin
      if FLife > 0 then
        Dec(FLife)
      else
        GameOver(False);
    end;

    edRed.Tag:= 3;
  end
  else if AEdit = edGreen then
  begin
    if edGreen.Text <> lbGreen.Text then
    begin
      if FLife > 0 then
        Dec(FLife)
      else
        GameOver(False);
    end;

    edRed.Tag:= 3;
  end
  else if AEdit = edBlue then
  begin
    if edBlue.Text <> lbBlue.Text then
    begin
      if FLife > 0 then
        Dec(FLife)
      else
        GameOver(False);
    end;

    edRed.Tag:= 3;
  end;
  UpdateLife;
end;

procedure TfrmMain.btnStartClick(Sender: TObject);
  function IsGameStart: Boolean;
  begin
    Result:= Sender = btnStart;
  end;

begin
  // A jtk kezdse/befejezse
  SwitchTo3D;
  try
    if not IsGameStart then
      EndGame;

    if Assigned(FContainer) then
    begin
      { 270 fokos forgats ngyzetes interpolcival }
      FContainer.AnimateFloatWait('RotationAngle.X', 270, 1.5, TAnimationType.atInOut, TInterpolationType.itQuadratic);
      if IsGameStart then
      begin
        { A jtktrre tvlts }
        tcMain.ActiveTab:= tiGame;
        ResetControls;
      end
      else
      begin
        { A foldalra tvlts }
        tcMain.ActiveTab:= tiMenu;
      end;
      { Tovbbforgats, hogy krberjen, de most "kicsit tlmegy-visszamegy" mdszerrel }
      FContainer.AnimateFloatWait('RotationAngle.X', 360, 0.5, TAnimationType.atInOut, TInterpolationType.itBack);
    end;

    if IsGameStart then
      StartGame;
  finally
    SwitchTo2D;
  end;
end;

procedure TfrmMain.ClearGameMatrix;
var
  I, J: Integer;
begin
  { Jtktr visszalltsa alapllapotba - sehol nincs semmi }
  for I := 1 to GAME_SIZE do
  begin
    for J := 1 to GAME_SIZE do
    begin
      FGameMatrix[I, J]:= False;
    end;
  end;
end;

procedure TfrmMain.edRedClick(Sender: TObject);
begin
  { Csak kattintsra kapja meg a fkuszt }
  (Sender as TEdit).Tag:= 1;
  (Sender as TEdit).SetFocus;
end;

procedure TfrmMain.edRedEnter(Sender: TObject);
begin
  { TAB-bal nem kaphat fkuszt a vezrl }
  if (Sender as TEdit).Tag <> 1 then
    btnEnd.SetFocus;
end;

procedure TfrmMain.edRedExit(Sender: TObject);
begin
  { TEdit.Tag jelentse:
     - 0: nem kaphat fkuszt
     - 1: kaphat fkuszt }
  (Sender as TEdit).Tag:= 0;
end;

procedure TfrmMain.SetGameLogic;
  procedure SetLogic(var ALogic: TEditLogic; ADelayTime: integer; AFadeInTime: Integer; AFadeInStep: Double;
    AFadeOutTime: Integer; AFadeOutStep: Double; AVisibleTime: Integer);
  begin
    ALogic.DelayTime:= ADelayTime;
    ALogic.FadeInTime:= AFadeInTime;
    ALogic.FadeInStep:= AFadeInStep;
    ALogic.FadeOutTime:= AFadeOutTime;
    ALogic.FadeOutStep:= AFadeOutStep;
    ALogic.VisibleTime:= AVisibleTime;
  end;

begin
  { A belltsok alapjn, a pontos nehzsgi tnyezk belltsa }
  case FDifficulty of
    1:
    begin
      SetLogic(FRedLogic, 1, 100, 0.2, 100, 0.2, 10000);
      SetLogic(FGreenLogic, 3000, 100, 0.2, 100, 0.2, 10000);
      SetLogic(FBlueLogic, 6000, 100, 0.2, 100, 0.2, 10000);
      FRotationDeltaMin:= -45;
      FRotationDeltaMax:= 45;
      FWordLength:= 4;
    end;
    2:
    begin
      SetLogic(FRedLogic, 1, 100, 0.2, 100, 0.2, 8000);
      SetLogic(FGreenLogic, 3000, 100, 0.2, 100, 0.2, 8000);
      SetLogic(FBlueLogic, 6000, 100, 0.2, 100, 0.2, 8000);
      FRotationDeltaMin:= 0;
      FRotationDeltaMax:= 90;
      FWordLength:= 5;
    end;
    3:
    begin
      SetLogic(FRedLogic, 1, 100, 0.2, 100, 0.2, 6000);
      SetLogic(FGreenLogic, 2000, 100, 0.2, 100, 0.2, 6000);
      SetLogic(FBlueLogic, 4000, 100, 0.2, 100, 0.2, 6000);
      FRotationDeltaMin:= 90;
      FRotationDeltaMax:= 180;
      FWordLength:= 6;
    end;
    4:
    begin
      SetLogic(FRedLogic, 1, 100, 0.2, 100, 0.2, 4000);
      SetLogic(FGreenLogic, 1000, 100, 0.2, 100, 0.2, 4000);
      SetLogic(FBlueLogic, 2000, 100, 0.2, 100, 0.2, 4000);
      FRotationDeltaMin:= 180;
      FRotationDeltaMax:= 270;
      FWordLength:= 7;
    end;
  end;
end;

procedure TfrmMain.SetGameMatrix(AControl: TControl; var X, Y: Integer);
var
  LX, LY: Integer;
begin
  { AControl elhelyezse a jtktren }
  if Assigned(AControl) then
  begin
    { Az elz hely felszabadul }
    LX:= Trunc(AControl.Position.X - 50 + Trunc(AControl.Width) div 2) div 100 + 1;
    LY:= Trunc(AControl.Position.Y - 50 + Trunc(AControl.Height) div 2) div 100 + 1;
    FGameMatrix[LX, LY]:= False;
  end;
  SetGameMatrix(X, Y);
end;

procedure TfrmMain.SetGameMatrix(var X: Integer; var Y: Integer);
begin
  { Egy pont elhelyezse a jtktren }
  repeat
    X:= 1 + Random(GAME_SIZE + 1 - 1);
    Y:= 1 + Random(GAME_SIZE + 1 - 1);
  until not FGameMatrix[X, Y];
  FGameMatrix[X, Y]:= True;
end;

procedure TfrmMain.StartGame;
begin
  { Jtk kezdetekor hvdik meg }
  FLife:= 5;
  SetGameLogic;
  FStartTime:= EncodeDateTime(2000, 1, 1, 0, FGameLength, 0, 0);

  UpdateTime;
  UpdateLife;
  UpdateEdit(edRed, FRedLogic, True);
  UpdateEdit(edGreen, FGreenLogic, True);
  UpdateEdit(edBlue, FBlueLogic, True);

  tmrTime.Enabled:= True;
  tmrRed.Enabled:= True;
  tmrGreen.Enabled:= True;
  tmrBlue.Enabled:= True;
end;

procedure TfrmMain.EndGame;
begin
  { Jtk befejezsekor hvdik meg, mieltt a menube tr vissza }
  tmrTime.Enabled:= False;
  tmrRed.Enabled:= False;
  tmrGreen.Enabled:= False;
  tmrBlue.Enabled:= False;
end;

procedure TfrmMain.FormCreate(Sender: TObject);
begin
  Randomize;
  FWordLength:= 3;
  FDifficulty:= 1;
  FGameLength:= 1;
  FLife:= 5;

  FRedLogic.Timer:= tmrRed;
  FGreenLogic.Timer:= tmrGreen;
  FBlueLogic.Timer:= tmrBlue;

  tcMain.ActiveTab:= tiMenu;
  { A tab flek elrejtse }
  tcMain.TabPosition:= FMX.TabControl.TTabPosition.tpNone;

  {$IFDEF DEBUG}
  memDebug.Visible:= True;
  {$ENDIF}

  adDifficultyChange(Self);
  adLengthChange(Self);
end;

procedure TfrmMain.SwitchTo3D;
begin
  { A 3D-s nzet s fellet ltrehozsa }
  FViewport := TViewport3D.Create(Self);
  FViewport.Parent := Self;
  FViewport.Align := TAlignLayout.alClient;
  FViewport.Color := claNull;
  FContainer := TLayer3D.Create(Self);
  FContainer.Parent := FViewport;
  FContainer.Projection := TProjection.pjScreen;
  FContainer.Align := TAlignLayout.alClient;
  { Az ablak mozgatsa a 3D-s szntrbe }
  layMain.Parent := FContainer;
end;

procedure TfrmMain.tmrRedTimer(Sender: TObject);
begin
  if Sender = tmrRed then
    UpdateEdit(edRed, FRedLogic)
  else if Sender = tmrGreen then
    UpdateEdit(edGreen, FGreenLogic)
  else if Sender = tmrBlue then
    UpdateEdit(edBlue, FBlueLogic);
end;

procedure TfrmMain.tmrTimeTimer(Sender: TObject);
begin
  FStartTime:= System.DateUtils.IncSecond(FStartTime, -1);
  UpdateTime;
  { Ha lejrt az id, akkor sikeresen vge a jtknak }
  if (Timeof(FStartTime) = 0) and (Minuteof(FStartTime) = 0) then
    GameOver(True);
end;


procedure TfrmMain.Generate(AEdit: TEdit; ALogic: TEditLogic; AIsInit: Boolean);
  procedure GenerateWord(ALabel: TLabel);
  var
    I: Integer;
    S: string;
  begin
    S:= '';
    for I := 1 to FWordLength do
    begin
      { A generlt szban angol kis abc beti lehetnek }
      S:= S + Char(97 + Random(122 - 97));
    end;
    ALabel.Text:= S;
  end;

  procedure RefreshMemText;
  var
    I, J: Integer;
    S: String;
  begin
    { A jtktr llapotnak kirsa }
    S:= '';
    for J := 1 to GAME_SIZE do
    begin
      for I := 1 to GAME_SIZE do
      begin
        if FGameMatrix[I, J] then
          S:= S + ' X '
        else
          S:= S + ' 0 ';
      end;
      S:= S + #13#10;
    end;
    memDebug.Text:= S;
  end;

var
  X, Y: Integer;

begin
  { Szvegmez generlsa a jtktrre }
  AEdit.Text:= '';
  if AIsInit then
    SetGameMatrix(nil, X, Y)
  else
    SetGameMatrix(AEdit, X, Y);

  AEdit.Position.X:= 100 * (X - 1) + 50 - (Trunc(AEdit.Width) div 2);
  AEdit.Position.Y:= 100 * (Y - 1) + 50 - (Trunc(AEDit.Height) div 2);
  AEdit.RotationAngle:= FRotationDeltaMin + Random(FRotationDeltaMax - FRotationDeltaMin);

  if AEdit = edRed then
  begin
    GenerateWord(lbRed);
  end
  else if AEdit = edGreen then
  begin
    GenerateWord(lbGreen);
  end
  else if AEdit = edBlue then
  begin
    GenerateWord(lbBlue);
  end;

  {@IFDEF DEBUG}
  RefreshMemText;
  {@ENDIF}
end;

procedure TfrmMain.ResetControls;
begin
  { Minden jtkban szerepl vezrl alapllapotra visszalltsa, jtk kezdetekor }
  edRed.Opacity:= 0;
  lbRed.Text:= '';
  edGreen.Opacity:= 0;
  lbGreen.Text:= '';
  edBlue.Opacity:= 0;
  lbBlue.Text:= '';
  lbTime.Text:= '00:00';
  lbLife.Text:= '?';
  ClearGameMatrix;
  lbStatus.Visible:= False;
  lbLifeText.Visible:= True;
  lbTimeText.Visible:= True;
  lbLife.Visible:= True;
  lbTime.Visible:= True;
end;

procedure TfrmMain.UpdateEdit(AEdit: TEdit; ALogic: TEditLogic; AIsInit: Boolean);
begin
  { Szvegmez llapotnak frisstse }
  if AIsInit then
  begin
    AEdit.Opacity:= 0;
    AEdit.RotationAngle:= 0;
    ALogic.Timer.Enabled:= False;
    ALogic.Timer.Interval:= ALogic.DelayTime;
    Generate(AEdit, ALogic, True);
    { Timer.Tag jelentse:
       0 - az edit lthatatlan (kezdeti llapot)
       1 - az edit lthatv vlik
       2 - az edit 100%osan lthat
       3 - az edit lthatatlann vlik }
    ALogic.Timer.Tag:= 0;
    Exit;
  end;

  case ALogic.Timer.Tag of
    0:
    begin
      ALogic.Timer.Tag:= 1;
      ALogic.Timer.Interval:= ALogic.FadeInTime;
    end;
    1:
    begin
      if AEdit.Opacity < 1 then
      begin
        AEdit.Opacity:= AEdit.Opacity + ALogic.FadeInStep;
        TGlowEffect(AEdit.Children[0]).BringToFront;
      end
      else
      begin
        AEdit.Opacity:= 1;
        ALogic.Timer.Tag:= 2;
        ALogic.Timer.Interval:= ALogic.VisibleTime;
      end;
    end;
    2:
    begin
      ALogic.Timer.Tag:= 3;
      ALogic.Timer.Interval:= ALogic.FadeOutTime;
    end;
    3:
    begin
      if AEdit.Opacity > 0 then
      begin
        AEdit.Opacity:= AEdit.Opacity - ALogic.FadeOutStep;
        TGlowEffect(AEdit.Children[0]).BringToFront;
      end
      else
      begin
        AEdit.Opacity:= 0;
        ALogic.Timer.Enabled:= False;
        { Sikerlt-e berni aa megfelel szveget? }
        CheckLife(AEdit);
        { Ha vge, akkor nem megy tovbb}
        if FLife <=0 then
          Exit;
        AEdit.Text:= '';
        Generate(AEdit, ALogic);
        ALogic.Timer.Enabled:= True;
        ALogic.Timer.Tag:= 1;
        ALogic.Timer.Interval:= ALogic.FadeInTime;
      end;
    end;
  end;
end;

procedure TfrmMain.UpdateLife;
begin
  { Htrelev prblkozsok kirsnak frisstse }
  lbLife.Text:= IntToStr(FLife);
end;

procedure TfrmMain.UpdateTime;
begin
  { Htralev id frisstse }
  lbTime.Text:= System.SysUtils.FormatDateTime('nn:ss', FStartTime);
end;

procedure TfrmMain.SwitchTo2D;
begin
  { Vissza 2D-be az ablak }
  layMain.Parent := Self;

  if Assigned(FContainer) then
    FreeAndNil(FContainer);
  if Assigned(FViewport) then
    FreeAndNil(FViewport);
end;

end.
